"""
Workspace to store images and data associated with a processing pipeline

This class is conceptually simple - you can store pretty much any data by setting
an attribute on the workspace and retrieve the resulting data as the same attribute.
The workspace is backed by a directory and image data is save to files rather than
store in memory.

This hides considerable complexity. Here are a few of the issues:

 - To ensure that image data really is kicked out of memory we create an ImageProxy
   object for each image. This simply stores the filename and metadata associated
   with an image. ``__getattribute__`` is overridden to return ImageProxy attributes
   as the underlying Image.

 - There is a special ImageProxy for an AslImage. This might go away if we can
   represent the full state of an AslImage using metadata alone.

"""

import os
import sys
import errno
import glob
import shutil
import tempfile

import six
import numpy as np
import pandas as pd
import yaml

from fsl.data.image import Image

from oxasl import AslImage
from oxasl.reporting import Report
from oxasl.utils import Tee

class ImageProxy(object):
    """
    Reference to a saved Image and it's metadata
    """
    def __init__(self, fname, md=None):
        self._fname = fname
        self._md = md

    def img(self):
        """
        Return an Image object by loading from the file
        """
        img = Image(self._fname, loadData=False)
        if self._md:
            for key, value in self._md:
                img.setMeta(key, value)
        return img

class AslImageProxy(ImageProxy):
    """
    Reference to a saved AslImage and it's metadata
    """

    def img(self):
        """
        Return an AslImage object from the file and stored metadata
        """
        return AslImage(self._fname, loadData=False, **self._md)

class Workspace(object):
    """
    A workspace for data processing

    The contents of a workspace are modelled as attributes on the workspace
    object. An initial set of contents can be specified using keyword arguments.
    For command line tools, typically these are provided directly from the
    OptionParser results, e.g.

        options, args = parser.parse_args(sys.argv)
        wsp = Workspace(**vars(options))

    A workspace is always associated with a physical directory. Certain types of
    objects are automatically be saved to the workspace. If no save directory
    is specified a temporary directory is created

    Supported types are currently:

         - ``fsl.data.image.Image`` - Saved as Nifti
         - 2D Numpy array - Saved as ASCII matrix

    All other attributes are serialized to YAML and stored in a special
    ``_oxasl.yml`` file.

    To avoid saving a particular item, use the ``add`` method rather than
    directly setting an attribute, as it supports a ``save`` option.
    """

    def __init__(self, savedir=None, input_wsp="input", parent=None, defaults=("corrected", "input"), auto_asldata=False, **kwargs):
        """
        Create workspace

        :param savedir: If specified, use this path to save data. Will be created
                        if it does not not already exist. If not specified a temporary
                        directory will be created so data can be flushed to disk rather
                        than held in memory.
        :param separate_input: If True a sub workspace named 'input' will be created
                               and initial data will be stored there. This sub workspace
                               will also be checked by default when attributes are
                               requested from the main workspace.
        :param auto_asldata: If True, automatically create an AslImage attribute
                             from the input keyword arguments
        :param log:     File stream to write log output to (default: sys.stdout)
        """
        # Have to set this first otherwise setattr fails!
        if savedir is not None:
            savedir = os.path.abspath(savedir)
            create_savedir = True
        else:
            savedir = tempfile.mkdtemp(prefix="oxasl_wsp")
            create_savedir = False
        self.set_item("savedir", savedir, save=False)

        self._parent = parent
        self._defaults = list(defaults)
        self._stuff = {}
        if create_savedir:
            if parent is not None:
                warn_overwrite = not parent._overwrite_warned
            else:
                warn_overwrite = True
            mkdir(savedir, log=self.ifnone("log", kwargs.get("log", sys.stdout)), 
                  warn_if_exists=warn_overwrite)
            self._overwrite_warned = True
    
        # Defaults - these can be overridden by kwargs but might be
        # already defined in parent workspace
        if "log" in kwargs or self.log is None:
            self.log = kwargs.pop("log", None)
            if self.log is None:
                logfile = open(os.path.join(self.savedir, "logfile"), "w")
                self.log = Tee(sys.stdout, logfile)
        if "debug" in kwargs or self.debug is None:
            self.debug = kwargs.pop("debug", False)
        if "log_cmds" in kwargs or self.log_cmds is None:
            self.log_cmds = kwargs.pop("log_cmds", False)
        if "log_cmdout" in kwargs or self.log_cmdout is None:
            self.log_cmdout = kwargs.pop("log_cmdout", False)
        if "report" in kwargs or self.report is None:
            self.report = kwargs.pop("report", Report())

        # Default log configuration for FSL wrapper commands
        if self.fsllog is None:
            self.fsllog = kwargs.pop("fsllog", None)
            if not self.fsllog:
                self.fsllog = {"stderr" : self.log}
                if self.debug or self.log_cmds:
                    self.fsllog.update({"cmd" : self.log})
                if self.debug or self.log_cmdout:
                    self.fsllog.update({"stdout" : self.log})

        # Set kwargs as attributes in input workspace (if configured)
        if input_wsp:
            input_wsp = self.sub(input_wsp)
        else:
            input_wsp = self

        for key, value in kwargs.items():
            setattr(input_wsp, key, value)

        # Auto-generate ASLImage object
        if auto_asldata:
            if kwargs.get("asldata", None) is None:
                raise ValueError("Input ASL file not specified\n")
            input_wsp.asldata = AslImage(self.asldata, **kwargs)

    def __getattribute__(self, name):
        ret = super(Workspace, self).__getattribute__(name)
        if isinstance(ret, ImageProxy):
            return ret.img()
        else:
            return ret

    def __getattr__(self, name):
        if name in self._defaults:
            return None

        ret = None
        for wsp in self._defaults:
            default_wsp = getattr(self, wsp)
            if isinstance(default_wsp, Workspace):
                val = getattr(default_wsp, name)
                if val is not None:
                    ret = val
                    break

        if ret is None and self._parent is not None:
            ret = getattr(self._parent, name)

        return ret

    def __setattr__(self, name, value):
        self.set_item(name, value)

    def ifnone(self, attr, alternative):
        """
        Return the value of an attribute, if set and not None, or
        otherwise the supplied alternative
        """
        ret = getattr(self, attr, None)
        if ret is None:
            ret = alternative
        return ret

    def set_item(self, name, value, save=True, save_name=None, save_fn=None):
        """
        Add an item to the workspace

        Normally this is achieved by assigning to an attribute
        directly, however this function exists to allow greater control
        where required.

        The item will be set as an attribute on the workspace
        If a save directory is configured and the value is a supported
        type it will be saved there. This can be disabled by setting save=False

        :param name: Name, must be a valid Python identifier
        :param value: Value to set
        :param save: If False, do not save item
        :param save_name: If specified, alternative name to use for saving this item
        :param save_fn: If specified, Callable which generates string representation of
                        value suitable for saving the item to a file
        """
        if save:
            if not save_name:
                save_name = name

            # Remove any existing file first - it could be left behind if
            # the extension is different or the new value is None
            existing_files = glob.glob(os.path.join(self.savedir, "%s.*" % save_name))
            if not isinstance(value, Workspace):
                existing_files += glob.glob(os.path.join(self.savedir, save_name))
            for existing_file in existing_files:
                if os.path.isdir(existing_file):
                    shutil.rmtree(existing_file)
                else:
                    os.remove(existing_file)

            if value is not None:
                if save_fn is not None:
                    with open(os.path.join(self.savedir, save_name), "w") as tfile:
                        tfile.write(save_fn(value))
                elif isinstance(value, Image):
                    # Save as Nifti file
                    fname = os.path.join(self.savedir, save_name)
                    value.save(fname)
                    value.name = save_name
                    # Replace images with ImageProxy objects to avoid excess in-memory storage
                    if isinstance(value, AslImage):
                        value = AslImageProxy(fname, md=dict(value.metaItems()))
                    elif isinstance(value, Image):
                        value = ImageProxy(fname, md=dict(value.metaItems()))

                elif isinstance(value, np.ndarray) and value.ndim in (1, 2):
                    # Save as ASCII matrix
                    with open(os.path.join(self.savedir, save_name + ".mat"), "w") as tfile:
                        tfile.write(matrix_to_text(value))
                elif not name.startswith("_") and isinstance(value, pd.DataFrame):
                    # Save data frame in CSV file
                    value.to_csv(os.path.join(self.savedir, save_name + ".csv"), index=True, header=True)
                elif not name.startswith("_") and isinstance(value, (int, float, six.string_types)):
                    # Save other attributes in JSON file
                    self._stuff[name] = value
                    self._save_stuff()

        super(Workspace, self).__setattr__(name, value)

    def sub(self, name, parent_default=True, **kwargs):
        """
        Create a sub-workspace, (i.e. a subdir of this workspace)

        This inherits the log configuration from the parent workspace. The savedir
        will be a subdirectory of the original workspace. Additional data may be
        set using keyword arguments. The child-workspace will be available
        as an attribute on the parent workspace.

        :param name: Name of sub workspace
        :param parent_default: If True, attribute values default to the parent workspace
                               if not set on the sub-workspace
        """
        savedir = os.path.join(self.savedir, name)
        if parent_default and name not in self._defaults:
            parent = self
        else:
            parent = None
            kwargs["log"] = self.log
            kwargs["debug"] = self.debug

        sub_wsp = Workspace(savedir=savedir, parent=parent, input_wsp=None, **kwargs)
        setattr(self, name, sub_wsp)
        return sub_wsp

    def _save_stuff(self):
        with open(os.path.join(self.savedir, "_oxasl.yml"), "w") as tfile:
            yaml.dump(self._stuff, tfile, default_flow_style=False)

def matrix_to_text(mat):
    """
    Convert matrix array to text using spaces/newlines as col/row delimiters
    """
    # Pad 1D array to 2D matrix
    if mat.ndim == 1:
        mat = np.array([mat,])
    rows = []
    for row in mat:
        rows.append(" ".join([str(v) for v in row]))
    return "\n".join(rows)

def text_to_matrix(text):
    """
    Convert space or comma separated file to matrix
    """
    fvals = []
    ncols = -1
    lines = text.splitlines()
    for line in lines:
        # Discard comments
        line = line.split("#", 1)[0].strip()
        # Split by commas or spaces
        vals = line.replace(",", " ").split()
        # Ignore empty lines
        if not vals:
            continue
        # Check correct number of columns
        if ncols < 0:
            ncols = len(vals)
        elif len(vals) != ncols:
            raise ValueError("File must contain a matrix of numbers with fixed size (rows/columns)")
        # Check all data is numeric
        for val in vals:
            try:
                float(val)
            except:
                raise ValueError("Non-numeric value '%s' found in matrix text" % val)
        fvals.append([float(v) for v in vals])
    return np.array(fvals)

def mkdir(dirname, fail_if_exists=False, warn_if_exists=True, log=sys.stdout):
    """
    Create a directory, including necessary subdirs
    """
    try:
        os.makedirs(dirname)
    except OSError as exc:
        if exc.errno == errno.EEXIST:
            if fail_if_exists:
                raise
            elif warn_if_exists:
                log.write("WARNING: mkdir - Directory %s already exists\n" % dirname)
    return os.path.abspath(dirname)



"""
Plugin system for OXASL
=======================

Design
~~~~~~

A plugin is registered as a Python entry point. It must be an instance of
OxaslPlugin and provide the following information:

 - Name
 - Brief description
 - Version
 - Optional OptionCategory defining command line options supported
 - Optional list of IAF values which the plugin supports modelling for
 - Optional PVC estimation method?
"""

from oxasl.options import OptionCategory

class OxaslPlugin:
    def __init__(self, name, description, version, **kwargs):
        self.name = name
        self.description = description
        self.version = version
        self.options = kwargs.get("options", OptionCategory())



"""
Tests for CALIB module

FIXME need satrecov tests
FIXME need sensitivity correction tests
FIXME need test with edge correction
"""
import math

import pytest
import numpy as np

from fsl.data.image import Image

from oxasl import Workspace, reg, AslImage, brain

def get_wsp():
    wsp = Workspace(debug=True)
    wsp.struc = Image(np.random.rand(10, 10, 10))
    wsp.calib = Image(np.random.rand(5, 5, 5))
    return wsp

def test_get_regfrom_supplied():
    """
    Test a pre-supplied regfrom is preferred to anything else
    """
    wsp = get_wsp()
    user_regfrom = Image(np.random.rand(5, 5, 5))
    wsp.regfrom = user_regfrom
    reg.get_regfrom(wsp)
    assert(np.allclose(user_regfrom.data, wsp.reg.regfrom.data))

def test_get_regfrom_asldata_mean_tc():
    """
    Test brain extracted ASL mean is used for TC data
    """
    wsp = get_wsp()
    wsp.asldata = AslImage(np.random.rand(5, 5, 5, 4), tis=[1, 2], iaf="tc", ibf="rpt")
    reg.get_regfrom(wsp)
    meanasl_brain = brain.brain(wsp, wsp.asldata.mean(), thresh=0.2)
    assert(np.allclose(meanasl_brain.data, wsp.reg.regfrom.data))

def test_get_regfrom_asldata_mean_ct():
    """
    Test brain extracted ASL mean is used for CT data
    """
    wsp = get_wsp()
    wsp.asldata = AslImage(np.random.rand(5, 5, 5, 4), tis=[1, 2], iaf="ct", ibf="rpt")
    reg.get_regfrom(wsp)
    meanasl_brain = brain.brain(wsp, wsp.asldata.mean(), thresh=0.2)
    assert(np.allclose(meanasl_brain.data, wsp.reg.regfrom.data))

def test_get_regfrom_calib():
    """
    Test brain extracted calibration image is used for differenced data
    """
    wsp = get_wsp()
    wsp.asldata = AslImage(np.random.rand(5, 5, 5, 4), tis=[1, 2], iaf="diff", ibf="rpt")
    reg.get_regfrom(wsp)
    calib_brain = brain.brain(wsp, wsp.calib, thresh=0.2)
    assert(np.allclose(calib_brain.data, wsp.reg.regfrom.data))



"""
Wrapper for epi_reg command
"""

import fsl.utils.assertions as asrt
from fsl.wrappers import wrapperutils  as wutils

@wutils.fileOrImage('epi', 't1', 't1brain', 'fmap', 'fmapmag', 'fmapmagbrain', 'gdc', 'wmseg', 'weight', outprefix='out')
@wutils.fileOrArray('init', 'out.mat')
@wutils.fslwrapper
def epi_reg(**kwargs):
    """Wrapper for the ``epi_reg`` command.
    
    Required options:
    
    :arg epi: EPI image
    :arg t1: T1 weighted image (wholehead)
    :arg t1_brain: T1 weighted image (brain extracted)
    :arg out: Output image name or LOAD to return in-memory image
    
    Additional options:

    :arg fmap:          fieldmap image (in rad/s)
    :arg fmapmag:       fieldmap magnitude image - wholehead extracted
    :arg fmapmagbrain:  fieldmap magnitude image - brain extracted
    :arg gdc:           Gradient-distortion corection warpfield
    :arg wmseg:         white matter segmentation of T1 image
    :arg echospacing:   Effective EPI echo spacing (sometimes called dwell time) - in seconds
    :arg pedir:         Phase encoding direction, dir = x/y/z/-x/-y/-z
    :arg weight:        weighting image (in T1 space)
    :arg nofmapreg:     do not perform registration of fmap to T1 (use if fmap already registered) 
    """

    valmap = {
        'nofmapreg' : wutils.SHOW_IF_TRUE,
    }

    cmd = ['epi_reg', ]
    cmd += wutils.applyArgStyle('--=', valmap=valmap, singlechar_args=True, **kwargs)
    
    return cmd



"""
Additional FSL wrappers intended to be compatible with FSL python wrappers as far as possible
"""

from .fabber import fabber, mvntool
from .epi_reg import epi_reg
from .fnirt_extra import fnirtfileutils

__all__ = ["fabber", "mvntool", "epi_reg", "fnirtfileutils"]



"""
For generating human-readable output reports

Example usage:

    report = Report()

    page = ReportPage()
    page.heading("Registration to standard space")
    page.heading("Structural image", level=1)
    page.text("This image was used as the registration source")

    page.heading("Standard image", level=1)
    page.text("This image was used to obtain the registration target")

    page.heading("Transformation", level=1)
    page.heading("Affine transformation structural->standard", level=2)
    page.maths()
    page.heading("Affine transformation standard->structural", level=2)
    page.maths()

    report.add("registration", page)
    report.generate_html("my-report")
"""
from __future__ import unicode_literals

import sys
import os
import math
import datetime
import shutil
import tempfile
import warnings
import csv

import six
import numpy as np
import scipy.ndimage

try:
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.figure import Figure
except ImportError:
    Figure = None

from fsl.data.image import Image

def which(program):
    """
    Simple implementation of which to search for executable
    """
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, _fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None

class LightboxImage(object):
    """
    A .png image file showing a lightbox view of one or more Image instances
    """

    def __init__(self, img, bgimage=None, mask=None, **kwargs):
        """
        :param img: ``fsl.data.Image`` instance
        :param bgimage: ``fsl.data.Image`` instance which will be used as a greyscale background
        :param mask: ``fsl.data.Image`` instance which will be used to mask ``img``

        Keyword arguments:

        :param zeromask: If True, treat zero values as transparent
        :param outline: If True, show image as an outline (assumes image is binarised)
        :param colorbar: If True, display colorbar
        """
        self._img = img
        self._bgimage = bgimage
        self._mask = mask
        self._zeromask = kwargs.get("zeromask", True)
        self._outline = kwargs.get("outline", False)
        self._colorbar = kwargs.get("colorbar", False)
        self._clamp_colors = kwargs.get("clamp_colors", True)
        self.extension = ".png"

    def _slicerange(self, img, shape):
        if img is not None:
            nonzero_slices = [idx for idx in range(shape[2]) if np.count_nonzero(img.data[:, :, idx]) > 0]
            if nonzero_slices:
                return min(nonzero_slices), max(nonzero_slices)
        else:
            return 0, shape[2]-1

    def tofile(self, fname):
        """
        Write image to a file
        """
        if Figure is None:
            warnings.warn("matplotlib not installed - cannot generate images")
            return

        shape = None
        for img in [self._img, self._bgimage, self._mask]:
            if img is None:
                continue
            if not isinstance(img, Image):
                raise ValueError("%s: Images must be instances of fsl.data.Image: %s" % (fname, img))
            if shape is None:
                shape = img.shape
            if img.ndim != 3:
                raise ValueError("Images must be 3D")
            if img.shape != shape:
                raise ValueError("Images do not have consistent shapes")

        min_slice, max_slice = self._slicerange(self._img, shape)
        num_slices = min(16, max_slice - min_slice + 1)
        grid_size = int(math.ceil(math.sqrt(num_slices)))

        fig = Figure(figsize=(5, 5), dpi=200)
        FigureCanvas(fig)
        for nslice in range(num_slices):
            axes = fig.add_subplot(grid_size, grid_size, nslice+1)
            axes.set_yticklabels([])
            axes.set_xticklabels([])
            axes.set_xticks([])
            axes.set_yticks([])
            slice_idx = int(float((max_slice - min_slice + 1)*nslice)/num_slices) + min_slice

            if self._bgimage:
                data = self._bgimage.data[:, :, slice_idx].T
                axes.imshow(data, cmap='gray')

            if self._img:
                data = self._img.data[:, :, slice_idx].T
                data[~np.isfinite(data)] = 0

                if issubclass(data.dtype.type, np.integer):
                    cmap = "Reds"
                    vmax, vmin = np.max(self._img.data), np.min(self._img.data)
                else:
                    cmap = "viridis"
                    vmax, vmin = np.percentile(self._img.data, 99), np.percentile(self._img.data, 1)
                    if vmax == vmin:
                        vmax, vmin = np.max(self._img.data), np.min(self._img.data)

                    if self._clamp_colors:
                        data = np.clip(data, vmin, vmax)

                if self._outline:
                    data = (data > 0.5).astype(np.int)
                    data = data - scipy.ndimage.morphology.binary_erosion(data, structure=np.ones((3, 3)))

                if self._mask:
                    data = np.ma.masked_array(data, self._mask.data[:, :, slice_idx].T == 0)
                elif self._zeromask:
                    data = np.ma.masked_array(data, data == 0)

                img = axes.imshow(data, vmax=vmax, vmin=vmin, cmap=cmap)

            # Reverse y-axis so anterior is as the top
            axes.set_ylim(axes.get_ylim()[::-1])

        fig.subplots_adjust(wspace=0, hspace=0.05)
        if self._img and self._colorbar:
            fig.colorbar(img, ax=fig.axes)

        fig.savefig(fname, bbox_inches='tight')

class LineGraph(object):
    """
    A .png image file showing a line graph
    """

    def __init__(self, data, xlabel, ylabel, **kwargs):
        """
        :param *imgs: One or more ``fsl.data.Image`` instances. Later images will be
                      overlaid onto earlier images
        :param zeromask: If True, treat zero values as transparent
        """
        self._data = data
        self._xlabel = xlabel
        self._ylabel = ylabel
        self._figsize = kwargs.get("figsize", (6, 3))
        self.extension = ".png"

    def tofile(self, fname):
        """
        Write image to a file
        """
        if Figure is None:
            warnings.warn("matplotlib not installed - cannot generate graphs")
            return

        fig = Figure(figsize=self._figsize, dpi=200)
        FigureCanvas(fig)
        axes = fig.add_subplot(1, 1, 1)
        axes.set_xlabel(self._xlabel)
        axes.set_ylabel(self._ylabel)
        axes.plot(self._data)

        fig.savefig(fname, bbox_inches='tight')

class ReportPage(object):
    """
    Simple helper class for creating documents containing ReStructuredText
    """

    def __init__(self, name, report, **kwargs):
        self._name = name
        self._report = report
        self._content = ""
        self._heading_chars = "=-~+"
        self.extension = ".rst"
        self._prepared = False

    def image(self, name, img_obj):
        """
        Add a block-level image
        """
        self._report.add(name, img_obj)
        self._content += ".. image:: %s%s\n\n" % (name, img_obj.extension)

    def text(self, txt):
        """
        Add a line of text content
        """
        self._content += txt + "\n\n"

    def maths(self, content):
        """
        Write mathematical content
        """
        self.text(".. math::")
        if isinstance(content, six.string_types):
            content = content.splitlines()

        for line in content:
            self._content += "    " + line + "\n"
        self._content += "\n"

    def matrix(self, mat, sig_fig=3):
        """
        Add a matrix of numbers
        """
        matrix_latex = "\\begin{bmatrix}\n"
        for row in mat:
            matrix_latex += " & ".join([self._latex_float(v, sig_fig) for v in row]) + " \\\\\n"
        matrix_latex += "\\end{bmatrix}\n"
        self.maths(matrix_latex)

    def heading(self, txt, level=0):
        """
        Add a heading
        """
        if level >= len(self._heading_chars):
            raise ValueError("Unsupported heading level: %i" % level)
        self._content += txt + "\n"
        self._content += self._heading_chars[level] * len(txt) + "\n\n"

    def table(self, tabdata, name="", headers=None):
        """
        Add a table
        """
        self._content += ".. csv-table:: " + name + "\n"
        if headers:
            self._content += "    :header: " + ",".join(['"%s"' % h for h in headers]) + "\n"
        self._content += "\n"

        csvtxt = six.StringIO()
        writer = csv.writer(csvtxt)
        # Required because csv does not support unicode in python 2
        if six.PY2:
            for row in tabdata:
                writer.writerow([unicode(s).encode("utf-8") for s in row])
            for line in csvtxt.getvalue().splitlines():
                self._content += "    " + line.decode("utf-8") + "\n"
        else:
            for row in tabdata:
                writer.writerow([str(s) for s in row])
            for line in csvtxt.getvalue().splitlines():
                self._content += "    " + line + "\n"

        self._content += "\n"

    def dicttable(self, dictionary):
        """
        Add table based on the contents of a python dictionary.

        The columns are 'Key' and 'Value'
        """
        tabdata = dictionary.items()
        self.table(tabdata, headers=("Key", "Value"))

    def tofile(self, fname):
        """
        Write RST content to a file
        """
        with open(fname, "wb") as rstfile:
            rstfile.write(self._content.encode("utf-8"))

    def _latex_float(self, val, sig_fig=3):
        """
        Format float in format suitable for Latex - nicked off StackOverflow!
        """
        pyformat = "{0:.%ig}" % sig_fig
        float_str = pyformat.format(val)
        if "e" in float_str:
            base, exponent = float_str.split("e")
            return r"{0} \times 10^{{{1}}}".format(base, int(exponent))
        else:
            return float_str

    def __str__(self):
        return self._content

class Report(object):
    """
    A report consisting of .rst documents and associated images
    which can be turned into a document (HTML, PDF etc)
    """

    def __init__(self, title="OXASL processing report"):
        self._contents = []
        self._files = {}
        self._start_time = datetime.datetime.now()
        self._end_time = None
        self.title = title
        self.extension = ""

    def generate_html(self, dest_dir, build_dir=None, log=sys.stdout):
        """
        Generate an HTML report
        """
        self._end_time = datetime.datetime.now()

        if build_dir:
            if os.path.exists(build_dir):
                if not os.path.isdir(build_dir):
                    raise ValueError("Report build directory %s exists but is not a directory" % build_dir)
                else:
                    warnings.warn("Report build directory %s already exists" % build_dir)
            is_temp = False
        else:
            build_dir = tempfile.mkdtemp("_report")
            is_temp = True

        try:
            self.tofile(build_dir)

            with open(os.path.join(build_dir, "conf.py"), "w") as conffile:
                conffile.write(REPORT_CONF)

            if os.path.exists(dest_dir):
                if not os.path.isdir(dest_dir):
                    raise ValueError("Report destination directory %s exists but is not a directory" % dest_dir)
                else:
                    warnings.warn("Report destination directory %s already exists - removing" % dest_dir)
                    shutil.rmtree(dest_dir)
            os.makedirs(dest_dir)

            try:
                # Different sphinx version have different main API.
                # We try the executable first currently to work around a bug in pyside2
                args = [which('sphinx-build'), '-b', 'html', build_dir, dest_dir]
                if args[0] is not None:
                    import subprocess
                    result = subprocess.call(args)
                else:
                    import sphinx
                    if hasattr(sphinx, "main"):
                        result = sphinx.main(args)
                    else:
                        import sphinx.cmd.build
                        result = sphinx.cmd.build.main(args[1:])
            except (AttributeError, ImportError, OSError):
                log.write("WARNING: sphinx not found, HTML report will not be generated\n")
                return False
            except Exception as exc:
                log.write("WARNING: sphinx failed, HTML report will not be generated\n")
                log.write("Message: %s\n" % str(exc))
                return False

            return True
        finally:
            if is_temp:
                shutil.rmtree(build_dir)

    def tofile(self, build_dir):
        """
        Write the report contents out to a build directory.

        This is the RST and image content - not the built HTML/PDF documentation
        """
        if not os.path.exists(build_dir):
            os.makedirs(build_dir)

        with open(os.path.join(build_dir, "index.rst"), "w") as indexfile:
            indexfile.write(self.title + "\n")
            indexfile.write("=" * len(self.title) + "\n\n")
            self._timings(indexfile)
            self._toc(indexfile)

        for fname, content in self._files.items():
            try:
                content.tofile(os.path.join(build_dir, fname))
            except Exception as exc:
                import traceback
                traceback.print_exc()
                warnings.warn("Error writing report content %s to file: %s" % (fname, exc))

    def page(self, name, overwrite=False, **kwargs):
        """
        Add a page to the report. The page is returned for content to be added

        :param name: Name of the page.
        :param overwrite: If True, and page already exists with the same ``name``,
                          replace content. Otherwise an exception is thrown.
        """
        page = ReportPage(name, report=self, **kwargs)
        self.add(name, page, overwrite)
        return page

    def add(self, name, content, overwrite=False):
        """
        Add content to a report

        :param name: Name of the content.
        :param content: Content object which has ``extension`` attribute and supports ``tofile()`` method
        :param overwrite: If True, and content already exists with the same ``name`` and extension,
                          replace content. Otherwise an exception is thrown.

        :return: Name given to the object in the report. If ``overwrite=True`` this will be the same as ``name``
        """
        fname = name + content.extension
        if not overwrite:
            idx = 2
            while fname in self._files:
                name = name + "_%i" % idx
                fname = name + content.extension

        self._files[fname] = content
        if isinstance(content, ReportPage):
            self._contents.append(name)
        if isinstance(content, Report):
            self._contents.append(name + "/index")
        return name

    def _timings(self, indexfile):
        if self._start_time:
            indexfile.write("Start time: %s\n\n" % self._start_time.strftime("%Y-%m-%d %H:%M:%S"))
        if self._end_time:
            indexfile.write("End time: %s\n\n" % self._end_time.strftime("%Y-%m-%d %H:%M:%S"))

    def _toc(self, indexfile):
        indexfile.write(".. toctree::\n")
        indexfile.write("  :maxdepth: 1\n")
        indexfile.write("  :caption: Contents:\n\n")
        for rst_file in self._contents:
            indexfile.write("  %s\n" % rst_file)

REPORT_CONF = """
# This file is execfile()d with the current directory set to its
# containing dir.

extensions = ['sphinx.ext.mathjax',]
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
project = u'oxasl'
copyright = u'2018, oxasl'
author = u'oxasl'
version = u''
release = u''
language = None
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']
pygments_style = 'sphinx'
todo_include_todos = False

html_theme = 'sphinxdoc'
# html_theme_options = {}
html_static_path = []
html_sidebars = {
    '**': [
        'relations.html',  # needs 'show_related': True theme option to display
        'searchbox.html',
    ]
}
htmlhelp_basename = 'oxasldoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # 'papersize': 'letterpaper',
    # 'pointsize': '10pt',
    # 'preamble': '',
    # 'figure_align': 'htbp',
}

latex_documents = [
    (master_doc, 'oxasl.tex', u'oxasl Documentation',
     u'oxasl', 'manual'),
]
"""

def main():
    """
    Simple command line for testing
    """
    report = Report()
    page = report.page("test")
    page.image("lbox", LightboxImage(*[Image(fname) for fname in sys.argv[1:]]))

    report.generate_html("testreport")

if __name__ == "__main__":
    main()



import pytest
import numpy as np

from fsl.data.image import Image

from oxasl import AslImage

def test_create_data_singleti():
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, tis=[1.5], iaf="tc", order="lrt")
    assert img.ntis == 1
    assert img.tis == [1.5]
    assert not img.have_plds
    assert img.rpts == [3]
    assert img.order == "lrt"

def test_create_data_numtis_single():
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, ntis=1, iaf="tc", order="lrt")
    assert img.ntis == 1
    assert not img.tis
    assert not img.have_plds
    assert img.rpts == [3]
    assert img.order == "lrt"

def test_create_data_multiti():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, tis=[1.5, 2.0], iaf="tc", order="lrt")
    assert img.ntis == 2
    assert img.tis == [1.5, 2.0]
    assert not img.have_plds
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_numtis_multi():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, ntis=2, iaf="tc", order="lrt")
    assert img.ntis == 2
    assert not img.tis
    assert not img.have_plds
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_multiti_var_repeats():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, tis=[1.5, 2.0], iaf="tc", order="lrt", rpts=[1, 3])
    assert img.ntis == 2
    assert img.tis == [1.5, 2.0]
    assert not img.have_plds
    assert img.rpts == [1, 3]
    assert img.order == "lrt"

def test_create_data_multiphase():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, tis=[1.5], iaf='mp', order='lrt', phases=[0, 45, 90, 135, 180, 225, 270, 315])
    assert img.ntis == 1
    assert img.tis == [1.5]
    assert not img.have_plds
    assert img.rpts == [1]
    assert img.ntc == 8
    assert img.phases == [0, 45, 90, 135, 180, 225, 270, 315]
    assert img.order == "lrt"

def test_create_data_multiphase_nphases():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, tis=[1.5], iaf="mp", order='lrt', nphases=8)
    assert img.ntis == 1
    assert img.tis == [1.5]
    assert not img.have_plds
    assert img.rpts == [1]
    assert img.ntc == 8
    assert img.phases == [0, 45, 90, 135, 180, 225, 270, 315]
    assert img.order == "lrt"

def test_create_data_plds():
    d = np.random.rand(5, 5, 5, 6)
    img = AslImage(name="asldata", image=d, plds=[1.5], tau=1.8, iaf="tc", order="lrt")
    assert img.ntis == 1
    assert img.plds == [1.5]
    assert img.tis == [pytest.approx(3.3)]
    assert img.have_plds
    assert img.casl
    assert img.rpts == [3]
    assert img.order == "lrt"

def test_create_data_multiplds():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, plds=[1.5, 1.6], tau=1.8, iaf="tc", order="lrt")
    assert img.ntis == 2
    assert img.plds == [1.5, 1.6]
    assert img.tis == [pytest.approx(3.3), pytest.approx(3.4)]
    assert img.have_plds
    assert img.casl
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_multitaus():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, plds=[1.5, 1.6], taus=[1.8, 2.0], iaf="tc", order="lrt")
    assert img.ntis == 2
    assert img.plds == [1.5, 1.6]
    assert img.tis == [pytest.approx(3.3), pytest.approx(3.6)]
    assert img.have_plds
    assert img.casl
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_strtaus():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, plds=[1.5, 1.6], taus="1.8, 2.0", iaf="tc", order="lrt")
    assert img.ntis == 2
    assert img.plds == [1.5, 1.6]
    assert img.tis == [pytest.approx(3.3), pytest.approx(3.6)]
    assert img.have_plds
    assert img.casl
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_floattaus():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, plds=[1.5, 1.6], tau=1.8, iaf="tc", order="lrt")
    assert img.ntis == 2
    assert img.plds == [1.5, 1.6]
    assert img.tis == [pytest.approx(3.3), pytest.approx(3.4)]
    assert img.have_plds
    assert img.casl
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_inttaus():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, plds=[1.5, 1.6], tau=1, iaf="tc", order="lrt")
    assert img.ntis == 2
    assert img.plds == [1.5, 1.6]
    assert img.tis == [pytest.approx(2.5), pytest.approx(2.6)]
    assert img.have_plds
    assert img.casl
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_plds_pasl():
    """ Odd usage, PLDs treated as TIs in non-CASL acquisitions """
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, plds=[1.5, 2.5], tau=1.8, casl=False, iaf="tc", order="lrt")
    assert img.ntis == 2
    assert img.tis == [1.5, 2.5]
    assert img.plds == [1.5, 2.5]
    assert img.have_plds
    assert img.rpts == [2, 2]
    assert img.order == "lrt"

def test_create_data_diff():
    d = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="asldata", image=d, tis=[1.5], iaf="diff", order='rt')
    assert img.ntis == 1
    assert img.tis == [1.5]
    assert not img.have_plds
    assert img.rpts == [8]
    assert img.ntc == 1
    assert img.order == "rt"

def test_diff_tc():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1], iaf="tc", order='lrt')
    img = img.diff()
    assert img.ntis == 1
    assert img.tis == [1]
    assert not img.have_plds
    assert img.rpts == [4]
    assert img.ntc == 1
    assert img.order == "rt"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 4] 
    assert np.all(data == 1)

def test_diff_ct():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1], iaf="ct", order='lrt')
    img = img.diff()
    assert img.ntis == 1
    assert img.tis == [1]
    assert not img.have_plds
    assert img.rpts == [4]
    assert img.ntc == 1
    assert img.order == "rt"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 4] 
    assert np.all(data == -1)

def test_reorder_tc_ct():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1], iaf="tc", order='lrt')
    img = img.reorder(iaf="ct")
    assert img.ntis == 1
    assert img.tis == [1]
    assert not img.have_plds
    assert img.rpts == [4]
    assert img.ntc == 2
    assert img.order == "lrt"
    assert img.iaf == "ct"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 8] 
    for z in range(8):
        if z % 2 == 0:
            assert np.all(data[..., z] == z+1)
        else:
            assert np.all(data[..., z] == z-1)

def test_reorder_lrt_rtl():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1], iaf="tc", order='lrt')
    img = img.reorder("rtl")
    assert img.ntis == 1
    assert img.tis == [1]
    assert not img.have_plds
    assert img.rpts == [4]
    assert img.ntc == 2
    assert img.order == "rtl"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 8] 
    for z in range(8):
        if z < 4:
            assert np.all(data[..., z] == z*2)
        else:
            assert np.all(data[..., z] == (z-4)*2+1)

def test_reorder_lrt_ltr():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], iaf="tc", order='lrt')
    img = img.reorder("ltr")
    assert img.ntis == 2
    assert img.tis == [1, 2]
    assert not img.have_plds
    assert img.rpts == [2, 2]
    assert img.ntc == 2
    assert img.order == "ltr"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 8] 
    for znew, zold in enumerate([0, 1, 4, 5, 2, 3, 6, 7]):
        assert np.all(data[..., znew] == zold)

def test_reorder_lrt_ltr_var_rpts():
    """ 
    This reordering with variable repeats is not supported. In priciple
    it is possible (TI1_R1, TI2_R1, TI2_R2, TI2_R3) but this seems unlikely
    and is probably more likely an error
    """
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], rpts = [1, 3], iaf="tc", order='lrt')
    with pytest.raises(Exception):
        img.reorder("ltr")
    #assert img.ntis == 2
    #assert img.tis == [1, 2]
    #assert not img.have_plds
    #assert img.rpts == [1, 3]
    #assert img.ntc == 2
    #assert img.order == "ltr"
    #data = img.nibImage.get_data()
    #assert list(data.shape) == [5, 5, 5, 8] 
    #for znew, zold in enumerate([0, 1, 2, 3, 4, 5, 6, 7]):
    #    assert np.all(data[..., znew] == zold)

def test_reorder_lrt_rlt():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], iaf="tc", order='lrt')
    img = img.reorder("rlt")
    assert img.ntis == 2
    assert img.tis == [1, 2]
    assert not img.have_plds
    assert img.rpts == [2, 2]
    assert img.ntc == 2
    assert img.order == "rlt"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 8] 
    for znew, zold in enumerate([0, 2, 1, 3, 4, 6, 5, 7]):
        assert np.all(data[..., znew] == zold)

def test_reorder_lrt_tlr():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], iaf="tc", order='lrt')
    img = img.reorder("tlr")
    assert img.ntis == 2
    assert img.tis == [1, 2]
    assert not img.have_plds
    assert img.rpts == [2, 2]
    assert img.ntc == 2
    assert img.order == "tlr"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 8] 
    for znew, zold in enumerate([0, 4, 1, 5, 2, 6, 3, 7]):
        assert np.all(data[..., znew] == zold)

def test_mean_across_repeats_rt():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='rt')
    img = img.mean_across_repeats()
    assert img.ntis == 2
    assert img.tis == [1, 2]
    assert not img.have_plds
    assert img.rpts == [1, 1]
    assert img.ntc == 1
    assert img.order == "rt"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 2]
    for znew, zold in enumerate([1.5, 5.5]):
        assert np.all(data[..., znew] == zold)

def test_mean_across_repeats_tr():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='tr')
    img = img.mean_across_repeats()
    assert img.ntis == 2
    assert img.tis == [1, 2]
    assert not img.have_plds
    assert img.rpts == [1, 1]
    assert img.ntc == 1
    assert img.order == "tr"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 2] 
    for znew, zold in enumerate([3, 4]):
        assert np.all(data[..., znew] == zold)

def test_mean_across_repeats_var_rpts():
    d = np.random.rand(5, 5, 5, 86)
    #for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2, 3, 4, 5], rpts=[6, 6, 6, 10, 15], iaf="tc", order='lrt')
    img = img.mean_across_repeats()
    assert img.ntis == 5
    assert img.tis == [1, 2, 3, 4, 5]
    assert img.rpts == [1, 1, 1, 1, 1]
    assert img.ntc == 1
    assert img.order == "rt"
    data = img.nibImage.get_data()
    assert list(data.shape) == [5, 5, 5, 5] 
    #for znew, zold in enumerate([3, 4]):
    #    assert np.all(data[..., znew] == zold)

def test_perf_weighted_tr():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='tr')
    pwi = img.perf_weighted()
    assert(isinstance(pwi, Image))
    data = pwi.data
    assert list(data.shape) == [5, 5, 5] 
    assert np.all(np.mean(d, -1) == data)

def test_vol_idx_tr():
    d = np.zeros([5, 5, 5, 8])
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='tr')
    assert img.get_vol_index(0, 0, 0) == 0
    assert img.get_vol_index(0, 1, 0) == 1
    assert img.get_vol_index(0, 0, 1) == 2
    assert img.get_vol_index(0, 1, 1) == 3
    assert img.get_vol_index(0, 0, 2) == 4
    assert img.get_vol_index(0, 1, 2) == 5
    assert img.get_vol_index(0, 0, 3) == 6
    assert img.get_vol_index(0, 1, 3) == 7
    
def test_vol_idx_rt():
    d = np.zeros([5, 5, 5, 8])
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='rt')
    assert img.get_vol_index(0, 0, 0) == 0
    assert img.get_vol_index(0, 1, 0) == 4
    assert img.get_vol_index(0, 0, 1) == 1
    assert img.get_vol_index(0, 1, 1) == 5
    assert img.get_vol_index(0, 0, 2) == 2
    assert img.get_vol_index(0, 1, 2) == 6
    assert img.get_vol_index(0, 0, 3) == 3
    assert img.get_vol_index(0, 1, 3) == 7
    
def test_vol_idx_lrt():
    d = np.zeros([5, 5, 5, 8])
    img = AslImage(name="asldata", image=d, tis=[1, 2], iaf="tc", order='lrt')
    assert img.get_vol_index(0, 0, 0) == 0
    assert img.get_vol_index(1, 0, 0) == 1
    assert img.get_vol_index(0, 1, 0) == 4
    assert img.get_vol_index(1, 1, 0) == 5
    assert img.get_vol_index(0, 0, 1) == 2
    assert img.get_vol_index(1, 0, 1) == 3
    assert img.get_vol_index(0, 1, 1) == 6
    assert img.get_vol_index(1, 1, 1) == 7
    
def test_vol_idx_ltr():
    d = np.zeros([5, 5, 5, 8])
    img = AslImage(name="asldata", image=d, tis=[1, 2], iaf="tc", order='ltr')
    assert img.get_vol_index(0, 0, 0) == 0
    assert img.get_vol_index(1, 0, 0) == 1
    assert img.get_vol_index(0, 1, 0) == 2
    assert img.get_vol_index(1, 1, 0) == 3
    assert img.get_vol_index(0, 0, 1) == 4
    assert img.get_vol_index(1, 0, 1) == 5
    assert img.get_vol_index(0, 1, 1) == 6
    assert img.get_vol_index(1, 1, 1) == 7
       
def test_vol_idx_var_rpts_rt():
    d = np.zeros([5, 5, 5, 7])
    img = AslImage(name="asldata", image=d, tis=[1, 2], rpts=[3, 4], iaf="diff", order='rt')
    assert img.get_vol_index(0, 0, 0) == 0
    assert img.get_vol_index(0, 0, 1) == 1
    assert img.get_vol_index(0, 0, 2) == 2
    assert img.get_vol_index(0, 1, 0) == 3
    assert img.get_vol_index(0, 1, 1) == 4
    assert img.get_vol_index(0, 1, 2) == 5
    assert img.get_vol_index(0, 1, 3) == 6
       
def test_vol_idx_var_rpts_tr():
    d = np.zeros([5, 5, 5, 7])
    img = AslImage(name="asldata", image=d, tis=[1, 2], rpts=[3, 4], iaf="diff", order='tr')
    assert img.get_vol_index(0, 0, 0) == 0
    assert img.get_vol_index(0, 1, 0) == 1
    assert img.get_vol_index(0, 0, 1) == 2
    assert img.get_vol_index(0, 1, 1) == 3
    assert img.get_vol_index(0, 0, 2) == 4
    assert img.get_vol_index(0, 1, 2) == 5
    assert img.get_vol_index(0, 1, 3) == 6
       
def test_vol_idx_var_rpts_lrt():
    d = np.zeros([5, 5, 5, 10])
    img = AslImage(name="asldata", image=d, tis=[1, 2], rpts=[3, 2], iaf="tc", order='lrt')
    assert img.get_vol_index(0, 0, 0) == 0
    assert img.get_vol_index(1, 0, 0) == 1
    assert img.get_vol_index(0, 0, 1) == 2
    assert img.get_vol_index(1, 0, 1) == 3
    assert img.get_vol_index(0, 0, 2) == 4
    assert img.get_vol_index(1, 0, 2) == 5
    assert img.get_vol_index(0, 1, 0) == 6
    assert img.get_vol_index(1, 1, 0) == 7
    assert img.get_vol_index(0, 1, 1) == 8
    assert img.get_vol_index(1, 1, 1) == 9
    
def test_split_epochs():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='tr')
    imgs = img.split_epochs(4)
    assert len(imgs) == 2
    for idx, img in enumerate(imgs):
        assert img.ntis == 2
        assert img.tis == [1, 2]
        assert not img.have_plds
        assert img.rpts == [1, 1]
        assert img.ntc == 1
        assert img.order == "tr"
        data = img.nibImage.get_data()
        # Epoch 1 is TIs 1212, data 0123, mean across repeats 12
        # Epoch 2 is TIs 1212, data 4567, mean across repeats 56
        start = idx*4
        for z in range(data.shape[3]):
            assert np.all(data[..., z] == start+z+1)

def test_split_epochs_overlap():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='tr')
    imgs = img.split_epochs(4, overlap=2)
    assert len(imgs) == 3
    for idx, img in enumerate(imgs):
        assert img.ntis == 2
        assert img.tis == [1, 2]
        assert not img.have_plds
        assert img.rpts == [1, 1]
        assert img.ntc == 1
        assert img.order == "tr"
        data = img.nibImage.get_data()
        # Epoch 1 is TIs 1212, data 0123, mean across repeats 12
        # Epoch 2 is TIs 1212, data 2345, mean across repeats 34
        # Epoch 3 is TIs 1212, data 4567, mean across repeats 56
        start = idx*2
        for z in range(data.shape[3]):
            assert np.all(data[..., z] == start+z+1)

def test_split_epochs_rt():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='rt')
    imgs = img.split_epochs(4)
    assert len(imgs) == 2
    for idx, img in enumerate(imgs):
        assert img.ntis == 1
        assert img.tis == [1+idx]
        assert not img.have_plds
        assert img.rpts == [1]
        assert img.ntc == 1
        assert img.order == "rt"
        data = img.nibImage.get_data()
        # Epoch 1 is TIs 1111, data 0123, mean across repeats 1.5
        # Epoch 2 is TIs 2222, data 4567, mean across repeats 5.5
        start = idx*4
        for z in range(data.shape[3]):
            assert np.all(data[..., z] == start+z+1.5)

def test_split_epochs_rt_overlap():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='rt')
    imgs = img.split_epochs(4, overlap=2)
    assert len(imgs) == 3
    for idx, img in enumerate(imgs):
        if idx in (0, 2):
            assert img.ntis == 1
            assert img.tis == [1+idx/2]
            assert not img.have_plds
            assert img.rpts == [1]
            assert img.ntc == 1
            assert img.order == "rt"
            data = img.nibImage.get_data()
            # Epoch 1 is TIs 1111, data 0123, mean across repeats 1.5
            # Epoch 3 is TIs 2222, data 4567, mean across repeats 5.5
            start = idx*2
            for z in range(data.shape[3]):
                assert np.all(data[..., z] == start+z+1.5)
        else:
            assert img.ntis == 2
            assert img.tis == [1, 2]
            assert not img.have_plds
            assert img.rpts == [1, 1]
            assert img.ntc == 1
            assert img.order == "rt"
            data = img.nibImage.get_data()
            # Epoch 2 is TIs 1122, data 2345, mean across repeats 2.5, 4.5
            for z in range(data.shape[3]):
                assert np.all(data[..., z] == 2*z+2.5)
            
def test_split_epochs_reorder():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='rt')
    imgs = img.split_epochs(4, time_order="tr")
    assert len(imgs) == 2
    for idx, img in enumerate(imgs):
        assert img.ntis == 2
        assert img.tis == [1, 2]
        assert not img.have_plds
        assert img.rpts == [1, 1]
        assert img.ntc == 1
        assert img.order == "tr"
        data = img.nibImage.get_data()
        # reordered = [0, 4, 1, 5, 2, 6, 3, 7]
        # Epoch 1 is TIs 1212, data 0415, mean across repeats 0.5, 4.5
        # Epoch 2 is TIs 1212, data 2637, mean across repeats 2.5, 6.5
        for z in range(data.shape[3]):
            assert np.all(data[..., z] == idx*2+0.5+z*4)
            
def test_split_epochs_reorder_overlap():
    d = np.zeros([5, 5, 5, 8])
    for z in range(8): d[..., z] = z
    img = AslImage(name="asldata", image=d, tis=[1, 2], order='rt')
    imgs = img.split_epochs(4, overlap=2, time_order="tr")
    assert len(imgs) == 3
    for idx, img in enumerate(imgs):
        assert img.ntis == 2
        assert img.tis == [1, 2]
        assert not img.have_plds
        assert img.rpts == [1, 1]
        assert img.ntc == 1
        assert img.order == "tr"
        data = img.nibImage.get_data()
        # reordered = [0, 4, 1, 5, 2, 6, 3, 7]
        # Epoch 1 is TIs 1212, data 0415, mean across repeats 0.5, 4.5
        # Epoch 2 is TIs 1212, data 1526, mean across repeats 1.5, 5.5
        # Epoch 3 is TIs 1212, data 2637, mean across repeats 2.5, 6.5
        for z in range(data.shape[3]):
            assert np.all(data[..., z] == idx+0.5+4*z)

def test_derived():
    d1 = np.random.rand(5, 5, 5, 8)
    d2 = np.random.rand(5, 5, 5, 8)
    img = AslImage(name="d1", image=d1, plds=[1, 2], iaf="ct", ibf="tis", casl=True, bolus=[3, 4])
    assert img.ntis == 2
    assert img.have_plds
    assert img.plds == [1, 2]
    assert img.tis == [4, 6]
    assert img.rpts == [2, 2]
    assert img.ntc == 2
    assert img.order == "lrt"
    
    img2 = img.derived(d2, name="d2")
    assert img2.ntis == 2
    assert img2.have_plds
    assert img2.plds == [1, 2]
    assert img2.tis == [4, 6]
    assert img2.rpts == [2, 2]
    assert img2.ntc == 2
    assert img2.order == "lrt"
    
def test_multite():
    d = np.random.rand(5, 5, 5, 16)
    img = AslImage(name="d", image=d, plds=[1, 2], iaf="tc", ibf="tis", tes=[8, 9], casl=True, bolus=[3, 4])
    assert img.ntes == 2
    assert img.tes == [8, 9]
    assert img.ntis == 2
    assert img.have_plds
    assert img.plds == [1, 2]
    assert img.tis == [4, 6]
    assert img.rpts == [2, 2]
    assert img.ntc == 2
    assert img.order == "elrt"
    assert img.nvols == 16

def test_multite_diff():
    d = np.zeros([5, 5, 5, 16])
    for z in range(16): d[..., z] = z
    img = AslImage(name="d", image=d, plds=[1, 2], iaf="tc", ibf="tis", tes=[8, 9], casl=True, bolus=[3, 4])
    imgdiff = img.diff()
    assert imgdiff.ntes == 2
    assert imgdiff.tes == [8, 9]
    assert imgdiff.ntis == 2
    assert imgdiff.have_plds
    assert imgdiff.plds == [1, 2]
    assert imgdiff.tis == [4, 6]
    assert imgdiff.rpts == [2, 2]
    assert imgdiff.ntc == 1
    assert imgdiff.order == "ert"
    assert imgdiff.nvols == 8
    np.all(imgdiff.data == 2)

def test_multite_mean_across_repeats_ibf_tis():
    d = np.zeros([5, 5, 5, 16])
    for z in range(16): d[..., z] = z
    img = AslImage(name="d", image=d, plds=[1, 2], iaf="tc", ibf="tis", tes=[8, 9], casl=True, bolus=[3, 4])
    imgmar = img.mean_across_repeats(diff=False)
    assert imgmar.ntes == 2
    assert imgmar.tes == [8, 9]
    assert imgmar.ntis == 2
    assert imgmar.have_plds
    assert imgmar.plds == [1, 2]
    assert imgmar.tis == [4, 6]
    assert imgmar.rpts == [1, 1]
    assert imgmar.ntc == 2
    assert imgmar.order == "elrt"
    assert imgmar.nvols == 8
    for z in range(8):
        if z < 4:
            # e.g. mean of 0th and 4th image = 2
            assert np.all(imgmar.data[..., z] == (z + 2))
        else:
            # e.g. mean of 8th and 12th image = 10
            assert np.all(imgmar.data[..., z] == (z + 6))

def test_multite_mean_across_repeats_ibf_rpt():
    d = np.zeros([5, 5, 5, 16])
    for z in range(16): d[..., z] = z
    img = AslImage(name="d", image=d, plds=[1, 2], iaf="tc", ibf="rpt", tes=[8, 9], casl=True, bolus=[3, 4])
    imgmar = img.mean_across_repeats(diff=False)
    assert imgmar.ntes == 2
    assert imgmar.tes == [8, 9]
    assert imgmar.ntis == 2
    assert imgmar.have_plds
    assert imgmar.plds == [1, 2]
    assert imgmar.tis == [4, 6]
    assert imgmar.rpts == [1, 1]
    assert imgmar.ntc == 2
    assert imgmar.order == "eltr"
    assert imgmar.nvols == 8
    for z in range(8):
        # e.g. mean of 0th and 8th image = 4
        assert np.all(imgmar.data[..., z] == (z + 4))

def test_calib_first_vol():
    """ Calibration image as first vol """
    d = np.random.rand(5, 5, 5, 9)
    img = AslImage(name="asldata", image=d, tis=[1.5, 2.0], iaf="tc", order="lrt", calib_first_vol=True)
    assert img.ntis == 2
    assert img.tis == [1.5, 2.0]
    assert not img.have_plds
    assert img.rpts == [2, 2]
    assert img.order == "lrt"
    assert np.all(img.data == d[..., 1:]) 
    assert img.calib is not None
    assert np.all(img.calib.data == d[..., 0])

def test_calib_first_vol_fail():
    """ Claim there is a calibration image as first vol but actually there isn't """
    d = np.random.rand(5, 5, 5, 8)
    with pytest.raises(Exception):
        img = AslImage(name="asldata", image=d, tis=[1.5, 2.0], iaf="tc", order="lrt", calib_first_vol=True)



"""
oxasl.gui.dist_corr_tab.py

Copyright (c) 2019 University of Oxford
"""
import wx

from oxasl.gui.widgets import TabPage

class AslDistCorr(TabPage):
    """
    Tab page containing options for distortion correction
    """

    FIELDMAP = 0
    CALIB_IMAGE = 1

    def __init__(self, parent, idx, n):
        TabPage.__init__(self, parent, "Distortion Correction", idx, n, name="distcorr")

        self.distcorr_choices = ["Fieldmap", "Calibration image"]

        self.section("Distortion Correction")

        # Calibration image options
        self.distcorr_cb = wx.CheckBox(self, label="Apply distortion correction")
        self.distcorr_cb.Bind(wx.EVT_CHECKBOX, self._changed)

        self.distcorr_ch = wx.Choice(self, choices=self.distcorr_choices[:1])
        self.distcorr_ch.SetSelection(self.FIELDMAP)
        self.distcorr_ch.Bind(wx.EVT_CHOICE, self._changed)
        self.pack("", self.distcorr_cb, self.distcorr_ch, enable=False)

        # Calib image options
        self.section("Calibration Image Mode")
        self.calib_picker = self.file_picker("Phase-encode-reversed calibration image")

        # Fieldmap options
        self.section("Fieldmap Mode")
        self.fmap_picker = self.file_picker("Fieldmap image (in rad/s)")
        self.fmap_mag_picker = self.file_picker("Fieldmap magnitude image")
        self.fmap_be_picker = self.file_picker("Brain-extracted magnitude image", optional=True)

        # General options
        self.section("General")
        self.echosp_num = self.number("Effective EPI echo spacing", minval=0, maxval=10)
        self.pedir_ch = self.choice("Phase encoding direction", choices=["x", "y", "z", "-x", "-y", "-z"])

        self.sizer.AddGrowableCol(1, 1)
        #sizer.AddGrowableRow(5, 1)
        self.SetSizer(self.sizer)
        self.next_prev()

    def options(self):
        options = {}
        if self.distcorr():
            options.update({
                "echospacing" : self.echosp_num.GetValue(),
                "pedir" : self.pedir_ch.GetStringSelection(),
            })
            if self.distcorr_fmap():
                options.update({
                    "fmap" : self.image("Fieldmap", self.fmap_picker.GetPath()),
                    "fmapmag" : self.image("Fieldmap magnitude", self.fmap_mag_picker.GetPath()),
                    "fmapmagbrain" : self.image("Fieldmap brain", self.fmap_be_picker.GetPath()),
                })
            else:
                options.update({
                    "cblip" : self.image("Phase-reversed calibration data", self.calib_picker.GetPath()),
                })
        return options

    def distcorr(self):
        """
        :return: True if distortion correction is enabled
        """
        return self.distcorr_cb.IsChecked()

    def distcorr_fmap(self):
        """
        :return: True if field distortion correction is selected
        """
        return self.distcorr_ch.GetSelection() == self.FIELDMAP

    def update(self):
        self.distcorr_ch.Enable(self.distcorr())

        cal = self.distcorr() and not self.distcorr_fmap()
        self.calib_picker.Enable(cal)

        fmap = self.distcorr() and self.distcorr_fmap()
        self.fmap_picker.Enable(fmap)
        self.fmap_mag_picker.Enable(fmap)
        self.fmap_be_picker.Enable(fmap)

        self.pedir_ch.Enable(self.distcorr())
        self.echosp_num.Enable(self.distcorr())

        TabPage.update(self)

    def calib_changed(self, enabled):
        """
        If calibration enabled, add the calibration image option for distortion correction
        """
        sel = self.distcorr_ch.GetSelection()
        if enabled:
            choices = self.distcorr_choices
            sel = 1
        else:
            choices = self.distcorr_choices[:1]
            sel = 0
        self.distcorr_ch.Enable(False)
        self.distcorr_ch.Clear()
        self.distcorr_ch.AppendItems(choices)
        self.distcorr_ch.SetSelection(sel)
        self.update()



"""
Misc utility functions
"""

import six

class Tee(object):
    """
    Output stream which keeps a string record of everything
    written to it, and also sends it on to any number of
    sub-streams
    """

    def __init__(self, *streams):
        self._streams = [six.StringIO(),]
        self._streams.extend(streams)

    def add(self, stream):
        """
        Add a sub-stream
        """
        if stream:
            self._streams.append(stream)

    def write(self, text):
        """
        Write to all output streams
        """
        for stream in self._streams:
            stream.write(text)

    def flush(self):
        """
        Flush all output streams
        """
        for stream in self._streams:
            stream.flush()

    def __str__(self):
        return self._streams[0].getvalue()



import numpy as np

from oxasl import Workspace, AslImage, preproc

def test_preproc_none():
    wsp = Workspace()
    d = np.random.rand(5, 5, 5, 6)
    wsp.asldata = AslImage(d, name="asldata", tis=[1.5], iaf="tc", order="lrt")

    preproc.preprocess(wsp)
    assert wsp.asldata.ntis == wsp.asldata_preproc.ntis
    assert wsp.asldata.tis == wsp.asldata_preproc.tis
    assert wsp.asldata.have_plds == wsp.asldata_preproc.have_plds
    assert wsp.asldata.rpts == wsp.asldata_preproc.rpts
    assert wsp.asldata.ntc == wsp.asldata_preproc.ntc
    assert wsp.asldata.order == wsp.asldata_preproc.order

def test_preproc_diff():
    wsp = Workspace()
    d = np.random.rand(5, 5, 5, 6)
    wsp.asldata = AslImage(d, name="asldata", tis=[1.5], iaf="tc", order="lrt")
    wsp.diff = True

    preproc.preprocess(wsp)
    assert wsp.asldata_preproc.ntis == wsp.asldata.ntis
    assert wsp.asldata_preproc.tis == wsp.asldata.tis
    assert wsp.asldata_preproc.have_plds == wsp.asldata.have_plds
    assert wsp.asldata_preproc.rpts == wsp.asldata.rpts
    assert wsp.asldata_preproc.ntc == 1
    assert wsp.asldata_preproc.order == "rt"

def test_preproc_reorder_diff():
    wsp = Workspace()
    d = np.random.rand(5, 5, 5, 6)
    wsp.asldata = AslImage(d, name="asldata", tis=[1.5], iaf="tc", order="lrt")
    wsp.diff = True
    wsp.reorder = "ltr"

    preproc.preprocess(wsp)
    assert wsp.asldata_preproc.ntis == wsp.asldata.ntis
    assert wsp.asldata_preproc.tis == wsp.asldata.tis
    assert wsp.asldata_preproc.have_plds == wsp.asldata.have_plds
    assert wsp.asldata_preproc.rpts == wsp.asldata.rpts
    assert wsp.asldata_preproc.ntc == 1
    assert wsp.asldata_preproc.order == "tr"

def test_preproc_diff_reorder():
    wsp = Workspace()
    d = np.random.rand(5, 5, 5, 6)
    wsp.asldata = AslImage(d, name="asldata", tis=[1.5], iaf="tc", order="lrt")
    wsp.diff = True
    wsp.reorder = "ltr"
    
    preproc.preprocess(wsp)
    assert wsp.asldata_preproc.ntis == wsp.asldata.ntis
    assert wsp.asldata_preproc.tis == wsp.asldata.tis
    assert wsp.asldata_preproc.have_plds == wsp.asldata.have_plds
    assert wsp.asldata_preproc.rpts == wsp.asldata.rpts
    assert wsp.asldata_preproc.ntc == 1
    assert wsp.asldata_preproc.order == "tr"

def test_preproc_smooth():
    wsp = Workspace()
    d = np.random.rand(5, 5, 5, 6)
    wsp.asldata = AslImage(d, name="asldata", tis=[1.5], iaf="tc", order="lrt")
    wsp.smooth = True
    wsp.fwhm = 1.5
    
    preproc.preprocess(wsp)
    assert wsp.asldata_preproc.ntis == wsp.asldata.ntis
    assert wsp.asldata_preproc.tis == wsp.asldata.tis
    assert wsp.asldata_preproc.have_plds == wsp.asldata.have_plds
    assert wsp.asldata_preproc.rpts == wsp.asldata.rpts
    assert wsp.asldata_preproc.ntc == wsp.asldata.ntc
    assert wsp.asldata_preproc.order == wsp.asldata.order

def test_preproc_moco():
    wsp = Workspace()
    d = np.random.rand(5, 5, 5, 6)
    wsp.asldata = AslImage(d, name="asldata", tis=[1.5], iaf="tc", order="lrt")
    wsp.mc = True
    
    preproc.preprocess(wsp)
    assert wsp.asldata_preproc.ntis == wsp.asldata.ntis
    assert wsp.asldata_preproc.tis == wsp.asldata.tis
    assert wsp.asldata_preproc.have_plds == wsp.asldata.have_plds
    assert wsp.asldata_preproc.rpts == wsp.asldata.rpts
    assert wsp.asldata_preproc.ntc == wsp.asldata.ntc
    assert wsp.asldata_preproc.order == wsp.asldata.order



"""
OXASL - Python package for ASL-MRI analysis
===========================================

Copyright (c) 2018 University of Oxford

The modules in this package are mostly Python replacements for existing shell script
and C++ code from the oxford_asl FSL module.

For many tools, FSL is required. The ``fslpy`` package is used to wrap required tools
such as BET and FAST.

Design
------

AslImage
~~~~~~~~
The ``oxasl.image.AslImage`` class represents a captured ASL data file. It
contains information about the acquisition (number/values of TIs/PLDs, repeats,
ordering of label/control image etc). It also has methods which act directly
on the data, for example performing tag-control subtraction, or generation of
a perfusion-weighted image.::

    img = AslImage("mpld.nii.gz", plds=[0.25, 0.5, 0.75, 1.0], iaf="tc", order='lrt')
    diffdata = img.diff()
    pwi = img.perf_weighted()

Workspaces
~~~~~~~~~~
The ``workspace`` module contains the ``Workspace`` class which
can be used to build a higher-level interface for complex workflows. A
workspace is simply an object whose attributes are images, data, etc being
used as part of the workflow. Unlike a normal object, requesting an attribute
that does not exist returns None rather than raising an exception.::

    wsp = Workspace()
    print(wsp.asldata) # prints None

A workspace can optionally be associated with a physical directory. If it is,
then setting attributes causes files to be saved in this directory for supported
data types, such as images or 2D matrices.::

    wsp = Workspace(savedir="myoutput")
    wsp.asldata = AslImage("mpld.nii.gz", ntis=1) # Saves myoutput/asldata.nii.gz

A workspace is also associated with
a log stream (sys.stdout by default) and a prepared logging dictionary ``fsllog``
for passing to FSL Python wrapper commands::

    wsp = Workspace()
    wsp.log.write("Hello World\n")
    wsp.rois.mask = fslmaths(img).bin().run(log=wsp.fsllog)

Module functions
~~~~~~~~~~~~~~~~
Other modules typically contains one or more functions which operate on a
workspace, in some cases with additional parameters (but not always).

Module functions operate under the general rule that data stored directly
as a workspace attribute is unprocessed, user-supplied data. Derived data
is then stored in a sub-workspace. Module functions will usually create
a sub-workspace to store their own output in, for example the ``struc`` module
places it's output (brain extractions, segmentations) in the wsp.structural
sub-workspace.

For example the ``calib`` module contains the ``calibrate`` function which
calibrates a perfusion image to physical units using either voxelwise or reference
region methods. It reads parameters required for this from the workspace, including
the calibration method to use.

Most of these functions write their output back into the workspace under a standard
name, however in some cases the function might be called on different input images
and might therefore return an image directly, which can be added to the workspace
by the caller under whatever name they prefer

Command line tools
~~~~~~~~~~~~~~~~~~
Most modules include a ``main()`` function which implements a command line
tool to wrap the module functionality. For example the ``preproc`` module
implements the ``oxasl_preproc`` command line tool which can be used to
do simple preprocessing of ASL data, such as the following to perform
label-control subtraction::

    oxasl_preproc -i asldata.nii.gz --nplds=5 --diff -o asldata_diff.nii.gz

Current ASL processing modules
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

 - :mod:`basil` - ASL Bayesian Model fitting using the Fabber code
 - :mod:`calib` - Calibration of perfusion data using voxelwise or reference region methods
 - :mod:`corrections` - Calculate and apply corrections (motion, distortion)
 - :mod:`mask` - Calculation of a suitable mask for brain data
 - :mod:`oxford_asl` - Unified processing pipeline for ASL brain data
 - :mod:`preproc` - Basic ASL preprocessing (label-control subtraction, etc)
 - :mod:`reg` - Registration between ASL, structural and standard spaces

Other modules
~~~~~~~~~~~~~

 - :mod:`image` - Definition of the main AslImage class
 - :mod:`reporting` - Generation of HTML reports from processing operations
 - :mod:`workspace` - Definition of the Workspace class
"""

try:
    from ._version import __version__, __timestamp__
except:
    __version__ = "unknown"
    __timestamp__ = "unknown"

from .image import AslImage
from .workspace import Workspace

# Work around ugly FSL log message
import logging
logging.basicConfig()
logging.getLogger("fsl.utils.platform").setLevel(logging.CRITICAL)

__all__ = [
    "__version__",
    "AslImage",
    "Workspace",
]



"""
Tests for workspace module
"""
import os
import shutil
import tempfile
from six import StringIO

import numpy as np
import pytest

from fsl.data.image import Image

from oxasl import Workspace, AslImage
from oxasl.workspace import text_to_matrix

def test_default_attr():
    """ Check attributes are None by default """
    wsp = Workspace()
    assert(wsp.wibble is None)

def test_set_attr():
    """ Check attributes can bet set """
    wsp = Workspace()
    assert(wsp.wibble is None)
    wsp.wibble = 7
    assert(wsp.wibble == 7)

def test_ctor_attributes():
    """ Check attributes specified in constructor """
    wsp = Workspace(wobble="hi")
    assert(wsp.wobble == "hi")

def test_log():
    """ Check that the log is picked up """
    log = StringIO()
    wsp = Workspace(log=log)
    wsp.log.write("hello")
    assert(log.getvalue() == "hello")

def test_ifnone():
    wsp = Workspace(wibble=11)
    assert(wsp.ifnone("wibble", 12) == 11)
    assert(wsp.ifnone("wobble", 12) == 12)

def test_sub():
    """ Test sub-workspaces """
    wsp = Workspace()
    wsp.sub("child")
    assert(isinstance(wsp.child, Workspace))
    assert(wsp.child.wibble is None)
    assert(wsp.child.log == wsp.log)
    
def test_sub_kwargs():
    """ Test creating a sub workspace with kwargs """
    wsp = Workspace()
    wsp.sub("child", wibble="squid", pudding=4)
    assert(isinstance(wsp.child, Workspace))
    assert(wsp.child.wibble == "squid")
    assert(wsp.child.pudding == 4)

def test_sub_inherit():
    """ Test sub workspaces can inherit values from their parent """
    wsp = Workspace()
    wsp.wibble = 7
    wsp.wobble = 6
    wsp.sub("child")
    wsp.child.wobble = 5
    assert(wsp.child.wibble == 7)
    assert(wsp.child.wobble == 5)

def test_sub_inherit_wsp():
    """ Test sub workspaces can inherit sub-workspaces from their parent """
    wsp = Workspace()
    wsp.sub("child1")
    wsp.child1.wibble = 7
    wsp.sub("child2")
    assert(wsp.child2.child1 is not None)
    assert(wsp.child2.child1.wibble == 7)

def test_input_wsp():
    """ Test putting constructor attributes in a default sub workspaces """
    wsp = Workspace(input_wsp="cakes", flapjack=4, fruit=3, defaults=[])
    assert(wsp.cakes is not None)
    assert(wsp.cakes.flapjack == 4)
    assert(wsp.cakes.fruit == 3)
    assert(wsp.flapjack is None)
    assert(wsp.fruit is None)

def test_default_wsp():
    """ Test default sub-workspaces for search """
    wsp = Workspace(defaults=["cars"])
    assert(wsp.cars is None)
    wsp.ferrari = 9
    wsp.merc = 8
    wsp.sub("cars")
    wsp.cars.porsche = 6
    wsp.cars.ferrari = 4
    assert(wsp.cars is not None)
    assert(wsp.ferrari == 9)
    assert(wsp.porsche == 6)
    assert(wsp.merc == 8)
    assert(wsp.cars.porsche == 6)
    assert(wsp.cars.ferrari == 4)
    assert(wsp.cars.merc is None)
    
def test_default_wsp_multiple():
    """ Test multiple default sub-workspaces for search """
    wsp = Workspace(defaults=["plants", "trees"])
    wsp.daffodil = 9
    wsp.larch = 1
    wsp.sub("trees")
    wsp.trees.oak = 3
    wsp.trees.larch = 2
    wsp.trees.apple = 7
    assert(wsp.daffodil == 9)
    assert(wsp.larch == 1)
    assert(wsp.oak == 3)
    assert(wsp.apple == 7)
    assert(wsp.trees.larch == 2)
    assert(wsp.trees.oak == 3)
    assert(wsp.trees.daffodil is None)
    assert(wsp.trees.apple == 7)
    wsp.sub("plants")
    wsp.plants.lily = 4
    wsp.plants.oak = 5
    assert(wsp.daffodil == 9)
    assert(wsp.larch == 1)
    assert(wsp.lily == 4)
    assert(wsp.oak == 5)
    assert(wsp.apple == 7)
    assert(wsp.trees.oak == 3)
    assert(wsp.trees.lily is None)
    assert(wsp.plants.daffodil is None)
    assert(wsp.plants.lily == 4)
    assert(wsp.plants.oak == 5)
    
def test_savedir_created():
    """ Test save dirs are created if they don't already exist """
    tempdir = tempfile.mktemp("_oxasl")
    try:
        log = StringIO()
        wsp = Workspace(savedir=tempdir, log=log)
        assert(wsp.savedir) == tempdir
        assert(os.path.isdir(tempdir))
        assert("WARNING" not in log.getvalue())
    finally:
        shutil.rmtree(tempdir)

def test_savedir_created_multilevel():
    """ Test multi-level save dirs are created if they don't already exist """
    tempdir = os.path.join(tempfile.mktemp("_oxasl"), "extra", "levels")
    try:
        log = StringIO()
        wsp = Workspace(savedir=tempdir, log=log)
        assert(wsp.savedir) == tempdir
        assert(os.path.isdir(tempdir))
        assert("WARNING" not in log.getvalue())
    finally:
        shutil.rmtree(tempdir)

def test_savedir_sub():
    """ Test sub-workspace have subdirs created """
    tempdir = tempfile.mktemp("_oxasl")
    try:
        log = StringIO()
        wsp = Workspace(savedir=tempdir, log=log)
        wsp.sub("quark")
        path = os.path.join(tempdir, "quark")
        assert(wsp.quark.savedir == path)
        assert(os.path.isdir(path))
        assert("WARNING" not in log.getvalue())
    finally:
        shutil.rmtree(tempdir)

def test_image_save():
    """ 
    Test images are saved in the savedir
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        img = Image(np.random.rand(5, 5, 5))
        wsp.testimg = img
        path = os.path.join(tempdir, "testimg.nii.gz")
        assert(os.path.isfile(path))
        otherimg = Image(path)
        assert(np.all(img.data == wsp.testimg.data))
        assert(np.all(img.data == otherimg.data))
    finally:
        shutil.rmtree(tempdir)

def test_image_nosave():
    """ 
    Test setting an image without saving
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        img = Image(np.random.rand(5, 5, 5))
        wsp.set_item("testimg", img, save=False)
        path = os.path.join(tempdir, "testimg.nii.gz")
        assert(not os.path.exists(path))
        assert(np.all(img.data == wsp.testimg.data))
    finally:
        shutil.rmtree(tempdir)

def test_image_save_name():
    """ 
    Test images are saved in the savedir with the specified name
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        img = Image(np.random.rand(5, 5, 5))
        wsp.set_item("testimg", img, save_name="pumpkin")
        path = os.path.join(tempdir, "testimg.nii.gz")
        assert(not os.path.exists(path))
        path = os.path.join(tempdir, "pumpkin.nii.gz")
        assert(os.path.isfile(path))
        otherimg = Image(path)
        assert(np.all(img.data == wsp.testimg.data))
        assert(np.all(img.data == otherimg.data))
    finally:
        shutil.rmtree(tempdir)

def test_matrix_save():
    """ 
    Test 2D matrices are saved in the savedir
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        mat = np.random.rand(4, 4)
        wsp.testmat = mat
        path = os.path.join(tempdir, "testmat.mat")
        assert(os.path.isfile(path))
        with open(path) as matfile:
            othermat = text_to_matrix(matfile.read())
        assert(np.all(mat == wsp.testmat))
        assert(np.all(mat == othermat))
    finally:
        shutil.rmtree(tempdir)

def test_matrix_nosave():
    """ 
    Test setting an matrix without saving
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        mat = np.random.rand(4, 4)
        wsp.set_item("testmat", mat, save=False)
        path = os.path.join(tempdir, "testmat.mat")
        assert(not os.path.exists(path))
        assert(np.all(mat == wsp.testmat))
    finally:
        shutil.rmtree(tempdir)

def test_matrix_save_name():
    """ 
    Test matrices are saved in the savedir with the specified name
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        mat = np.random.rand(4, 4)
        wsp.set_item("testmat", mat, save_name="parsnip")
        path = os.path.join(tempdir, "testmat.mat")
        assert(not os.path.exists(path))
        path = os.path.join(tempdir, "parsnip.mat")
        assert(os.path.isfile(path))
        with open(path) as matfile:
            othermat = text_to_matrix(matfile.read())
        assert(np.all(mat == wsp.testmat))
        assert(np.all(mat == othermat))
    finally:
        shutil.rmtree(tempdir)

def _custom_save(mat):
    return "Custom Save"

def test_custom_save():
    """ 
    Test matrices are saved in the savedir with the specified name
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        mat = np.random.rand(4, 4)
        wsp.set_item("testmat", mat, save_fn=_custom_save)
        path = os.path.join(tempdir, "testmat")
        assert(os.path.exists(path))
        with open(path) as sfile:
            assert("Custom Save" == sfile.read())
    finally:
        shutil.rmtree(tempdir)

def test_custom_save_name():
    """ 
    Test matrices are saved in the savedir with the specified name
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        mat = np.random.rand(4, 4)
        wsp.set_item("testmat", mat, save_name="potato", save_fn=_custom_save)
        path = os.path.join(tempdir, "testmat")
        assert(not os.path.exists(path))
        path = os.path.join(tempdir, "potato")
        assert(os.path.exists(path))
        with open(path) as sfile:
            assert("Custom Save" == sfile.read())
    finally:
        shutil.rmtree(tempdir)

def test_custom_save_nosave():
    """ 
    Test matrices are saved in the savedir with the specified name
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        wsp = Workspace(savedir=tempdir)
        mat = np.random.rand(4, 4)
        wsp.set_item("testmat", mat, save_fn=_custom_save, save=False)
        path = os.path.join(tempdir, "testmat")
        assert(not os.path.exists(path))
    finally:
        shutil.rmtree(tempdir)

def test_savedir_already_exists():
    """ 
    Test warning when save dir already exists
    """
    tempdir = tempfile.mkdtemp("_oxasl")
    try:
        log = StringIO()
        wsp = Workspace(savedir=tempdir, log=log)
        assert("WARNING" in log.getvalue())
        assert("already exists" in log.getvalue())
    finally:
        shutil.rmtree(tempdir)

def test_fsllog_default():
    """
    Test the FSL logging context created
    """
    log = StringIO()
    wsp = Workspace(log=log)
    assert(isinstance(wsp.fsllog, dict))
    assert(wsp.fsllog.get("stdout", None) is None)
    assert(wsp.fsllog.get("stderr", None) == log)
    assert(wsp.fsllog.get("cmd", None) is None)

def test_fsllog_debug():
    """
    Test the FSL logging context created in debug mode
    """
    log = StringIO()
    wsp = Workspace(debug=True, log=log)
    assert(isinstance(wsp.fsllog, dict))
    assert(wsp.fsllog.get("stdout", None) == log)
    assert(wsp.fsllog.get("stderr", None) == log)
    assert(wsp.fsllog.get("cmd", None) == log)

def test_aslimage():
    kwargs = {
        "asldata" : np.random.rand(5, 5, 5, 8),
        "tis" : [1, 2],
        "iaf" : "tc",
        "ibf" : "rpt",
    }
    wsp = Workspace(auto_asldata=True, **kwargs)
    assert(isinstance(wsp.asldata, AslImage))
    assert(wsp.asldata.tis == [1, 2])
    assert(wsp.asldata.iaf == "tc")
    assert(wsp.asldata.order == "ltr")
    assert(wsp.asldata.rpts == [2, 2])
    
def test_aslimage_missing():
    with pytest.raises(ValueError):
        Workspace(auto_asldata=True)
    with pytest.raises(ValueError):
        Workspace(auto_asldata=True, asldata=None)

def test_text_to_matrix_spaces():
    """
    Check that text_to_matrix works with space separated data
    """
    text = "1 2 3\n4 5 6\n"
    mat = text_to_matrix(text)
    assert(np.all(mat == [[1, 2, 3], [4, 5, 6]]))

def test_text_to_matrix_comma():
    """
    Check that text_to_matrix works with comma separated data
    """
    text = "1, 2, 3\n4,5,6\n"
    mat = text_to_matrix(text)
    assert(np.all(mat == [[1, 2, 3], [4, 5, 6]]))
    
def test_text_to_matrix_tabs():
    """
    Check that text_to_matrix works with tab separated data
    """
    text = "1\t2\t3\n4\t 5\t 6\n"
    mat = text_to_matrix(text)
    assert(np.all(mat == [[1, 2, 3], [4, 5, 6]]))
        
def test_text_to_matrix_mixed():
    """
    Check that text_to_matrix works with mixed separators
    """
    text = "1\t2   3\n4  ,  5, \t 6\n"
    mat = text_to_matrix(text)
    assert(np.all(mat == [[1, 2, 3], [4, 5, 6]]))
    
def test_text_to_matrix_not_matrix():
    text = "1 2 3\n4 5\n"
    with pytest.raises(ValueError):
        mat = text_to_matrix(text)
    
def test_text_to_matrix_not_numbers():
    text = "1 x 3\n4 5 6\n"
    with pytest.raises(ValueError):
        mat = text_to_matrix(text)
